
-------------------------------------------------------------------------------
-- Optimization
-------------------------------------------------------------------------------

-- GLOBALS: _G, CHAT_FRAMES, ConvertRGBtoColorString, FONT_COLOR_CODE_CLOSE, format, math, SixUI_ChatFrame, string, table, tonumber, UISpecialFrames, UNKNOWN, YELLOW_FONT_COLOR_CODE

-------------------------------------------------------------------------------
-- Namespace
-------------------------------------------------------------------------------

local A, B, C, D = unpack(select(2, ...)) -- AddOn, Border, Constants, Database
local M = A:NewModule("Chat Copy")
local Social = LibStub("SixSocial-1.0")

local CopyBox = SixUI_CopyBox
local ScrollFrame = CopyBox.scrollFrame
local EditBox = ScrollFrame.editBox

-------------------------------------------------------------------------------
-- Module
-------------------------------------------------------------------------------

-- Balances color codes (|c...) and escape sequences (|r).
function M.BalanceColorCodes(inputString)
	local colorSeqCount, escapeSeqCount = 0, 0

	for colorSeq, escapeSeq in string.gmatch(inputString, "(|c%x%x%x%x%x%x%x%x)(|r?)") do
		colorSeqCount = colorSeqCount + (colorSeq and 1 or 0)
		escapeSeqCount = escapeSeqCount + (escapeSeq == "|r" and 1 or 0)
	end

	return inputString .. string.rep("|r", math.max(0, colorSeqCount - escapeSeqCount))
end

-- Formats a BattleTag for display.
function M.FormatBattleTag(accountID)
	local battleTag = (Social:GetFriendByAccountID(tonumber(accountID)) or {}).battleTag
	return format("[%s]", string.split("#", battleTag or UNKNOWN))
end

-- Formats a numeric value with the appropriate singular or plural text label.
function M.FormatPlural(value, singular, plural)
	return format("%s %s", value, value == "1" and singular or plural)
end

function M:OnInitialize()
	table.insert(UISpecialFrames, "SixUI_CopyBox")
	EditBox:SetFontObject(C.TEXT_MEDIUM)
end

function M:OnEnable()
	if not self:IsHooked(SixUI_ChatFrame, "OnMouseDown") then
		self:HookScript(SixUI_ChatFrame, "OnMouseDown")
	end

	self:RegisterMessage("SIXUI_COPY_LINES")
end

function M:Copy(lines)
	if CopyBox:IsShown() then
		CopyBox:Hide()
	else
		EditBox:SetText(table.concat(lines, "\n"))
		CopyBox:Show()
	end
end

function M:GetShownChatFrame()
	for index = 1, #CHAT_FRAMES do
		local ChatFrame = _G["ChatFrame" .. index]

		if ChatFrame:IsShown() then
			return ChatFrame
		end
	end
end

function M:GetLines(frame)
	local lines = {}

	for index = 1, frame:GetNumMessages() do
		local text, r, g, b = frame:GetMessageInfo(index)
		local colorCode = r and g and b and ConvertRGBtoColorString({r = r, g = g, b = b}) or YELLOW_FONT_COLOR_CODE

		text = format("%s%s|r", colorCode, text:gsub(FONT_COLOR_CODE_CLOSE, FONT_COLOR_CODE_CLOSE .. colorCode))
		text = self.BalanceColorCodes(text)
		text = text:gsub("|HBNplayer:[^:]-:(%d+):.-|h.-|h", self.FormatBattleTag)
		text = text:gsub("|K[^|]+|k", "<" .. UNKNOWN .. ">")
		text = text:gsub("|TInterface.-|t", "")
		text = text:gsub("(%S-) |4(%w+):(%w+);", self.FormatPlural)

		table.insert(lines, text)
	end

	return lines
end

-------------------------------------------------------------------------------
-- Events
-------------------------------------------------------------------------------

function M:SIXUI_COPY_LINES(event, lines)
	self:Copy(lines)
end

-------------------------------------------------------------------------------
-- Hooks
-------------------------------------------------------------------------------

function M:OnMouseDown(frame, button)
	if (button ~= "RightButton") then return end
	local ChatFrame = self:GetShownChatFrame()
	if (not ChatFrame) then return end
	local lines = self:GetLines(ChatFrame)
	self:Copy(lines)
end
