
-------------------------------------------------------------------------------
-- Optimization
-------------------------------------------------------------------------------

-- GLOBALS: CLEAR_ALL, ERR_NO_REPLY_TARGET, GetServerTime, HISTORY, next, rawget, select, setmetatable, string, table

-------------------------------------------------------------------------------
-- Namespace
-------------------------------------------------------------------------------

local A, B, C, D = unpack(SixUI) -- AddOn, Border, Constants, Database
local M = A:NewModule("Display History")
local O = LibStub("LibDataBroker-1.1"):NewDataObject("SixUI History", {
	type = "data source",
	text = UNKNOWN,
	label = "History",
	glyph = C.Enum.Glyph.CommentSlash,
})
local Social = LibStub("SixSocial-1.0")
local Tooltip = LibStub("SixTooltip-1.0"):CreateTooltip({
	columnJustification = {"LEFT"},
	font = C.TEXT_REGULAR,
	headerFont = C.TEXT_LARGE,
	key = "SixUI History",
	numColumns = 1,
	separatorHeight = C.TOOLTIP_SEPARATOR_HEIGHT,
})
local Tooltip2 = LibStub("SixTooltip-1.0"):CreateTooltip({
	columnJustification = {"LEFT"},
	font = C.TEXT_REGULAR,
	headerFont = C.TEXT_LARGE,
	key = "SixUI History Menu",
	numColumns = 1,
	separatorHeight = C.TOOLTIP_SEPARATOR_HEIGHT,
	skipReleaseAllTooltips = true,
})

local chats = {}
local showingList = false

M.Enum = {}

M.Enum.Chat = {
	AccountId = "accountID",
	BattleTag = "battleTag",
	Label = "label",
	LastMessage = "lastMessage",
	Messages = "messages",
	Type = "type",
	Unit = "unit",
}

M.Enum.ChatType = {
	BattleNet = 0,
	Character = 1,
}

M.Enum.Message = {
	Read = "read",
	Text = "text",
	Time = "time",
	Type = "type",
}

M.Enum.MessageType = {
	Sent = 0,
	Received = 1,
}

local CHAT_METATABLE = {
	__index = function (t, k)
		if k == "lastMessage" then
			return t.messages[#t.messages]
		else
			return rawget(t, k)
		end
	end,
	__lt    = function (a, b)
		if a.lastMessage and b.lastMessage then
			return a.lastMessage.time > b.lastMessage.time
		else
			return not a.lastMessage
		end
	end,
}

local CHAT_TYPE = {
	CHAT_MSG_BN_WHISPER = M.Enum.ChatType.BattleNet,
	CHAT_MSG_BN_WHISPER_INFORM = M.Enum.ChatType.BattleNet,
	CHAT_MSG_WHISPER = M.Enum.ChatType.Character,
	CHAT_MSG_WHISPER_INFORM = M.Enum.ChatType.Character,
}

local COLOR = {
	[M.Enum.ChatType.BattleNet] = {
		[M.Enum.MessageType.Sent] = LIGHTGRAY_FONT_COLOR,
		[M.Enum.MessageType.Received] = CreateColor(0, 1, 0.965),
	},
	[M.Enum.ChatType.Character] = {
		[M.Enum.MessageType.Sent] = LIGHTGRAY_FONT_COLOR,
		[M.Enum.MessageType.Received] = CreateColor(1, 0.5, 1),
	},
}

local COPY_FORMAT = "[W %s] [%s] %s"

local MESSAGE_METATABLE = {
	__lt = function (a, b) return (a.time < b.time) end,
}

local MESSAGE_TYPE = {
	CHAT_MSG_BN_WHISPER = M.Enum.MessageType.Received,
	CHAT_MSG_BN_WHISPER_INFORM = M.Enum.MessageType.Sent,
	CHAT_MSG_WHISPER = M.Enum.MessageType.Received,
	CHAT_MSG_WHISPER_INFORM = M.Enum.MessageType.Sent,
}

local PREFIX = {
	[M.Enum.MessageType.Sent] = "To",
	[M.Enum.MessageType.Received] = "From",
}

-------------------------------------------------------------------------------
-- Helpers
-------------------------------------------------------------------------------

local function Chat_OnEnter(self, chatIndex)
	Tooltip2:Show(self, chats[chatIndex])
end

local function Chat_OnLeave(self)
	Tooltip2:Hide()
end

local function Chat_OnClick(self, chatIndex, button)
	local chat = chats[chatIndex]
	local friend = Social:GetFriendByAccountID(chat.accountID) or Social:WrapUnitInFriend(chat.unit)

	A:HandleClick({
		{mouseButton = C.Enum.MouseButton.Left,  modifierKey = C.Enum.ModifierKey.Control, callback = function () M:DeleteChat(chatIndex) end},
		{mouseButton = C.Enum.MouseButton.Left,  modifierKey = C.Enum.ModifierKey.Alt,     callback = function () if friend then friend:SendInvite() end end},
		{mouseButton = C.Enum.MouseButton.Left,  modifierKey = C.Enum.ModifierKey.Shift,   callback = function () if friend then friend:SendWho() end end},
		{mouseButton = C.Enum.MouseButton.Left,  modifierKey = C.Enum.ModifierKey.None,    callback = function () if friend then friend:SendTell() end end},
		{mouseButton = C.Enum.MouseButton.Right, modifierKey = C.Enum.ModifierKey.None,    callback = function () M:CopyChat(chat) end},
	}, button)
end

local function ClearAll_OnClick(self, button)
	M:TruncateChats(0)
end

local function GetChatLabel(battleTag, unit)
	return battleTag and string.split("#", battleTag) or A.FormatUnit(unit)
end

-------------------------------------------------------------------------------
-- Module
-------------------------------------------------------------------------------

function M:OnInitialize()
	self.db = D.db:RegisterNamespace(self:GetName(), {global = {chats = {}}})

	chats = self.db.global.chats

	for _, chat in next, chats do
		setmetatable(chat, CHAT_METATABLE)

		for _, message in next, chat.messages do
			setmetatable(message, MESSAGE_METATABLE)
		end
	end
end

function M:OnEnable()
	self:RegisterEvent("CHAT_MSG_BN_WHISPER")
	self:RegisterEvent("CHAT_MSG_BN_WHISPER_INFORM", "CHAT_MSG_BN_WHISPER")
	self:RegisterEvent("CHAT_MSG_WHISPER")
	self:RegisterEvent("CHAT_MSG_WHISPER_INFORM", "CHAT_MSG_WHISPER")
	self:RegisterMessage("SIXSOCIAL_BNFRIEND_UPDATE", "OnEvent")
	self:RegisterMessage("SIXUI_HISTORY_CHAT_CREATED", "OnEvent")
	self:RegisterMessage("SIXUI_HISTORY_CHAT_DELETED", "OnEvent")
	self:RegisterMessage("SIXUI_HISTORY_CHATS_RESTORED", "OnEvent")
	self:RegisterMessage("SIXUI_HISTORY_MESSAGE_CREATED", "OnEvent")
	self:RegisterMessage("SIXUI_HISTORY_MESSAGE_READ", "OnEvent")
	self:RestoreChats()
end

function M:OnEvent(event, ...)
	local chat, message = ...

	if event == "SIXSOCIAL_BNFRIEND_UPDATE" then
		self:RestoreChats()
	elseif event == "SIXUI_HISTORY_MESSAGE_CREATED" then
		table.insert(chat.messages, message)
		self:TruncateMessages(chat.messages)
		self:SortChats()
	elseif event == "SIXUI_HISTORY_CHAT_CREATED" then
		table.insert(chats, chat)
		self:SortChats()
		self:TruncateChats(25)
	elseif event == "SIXUI_HISTORY_MESSAGE_READ" then
		message.read = GetServerTime()
	end

	O:Update()

	if event ~= "SIXUI_HISTORY_MESSAGE_READ" then
		Tooltip:Refresh()
	end
end

function M:RestoreChats()
	for _, chat in next, chats do
		self:RestoreChat(chat)
	end

	self:SendMessage("SIXUI_HISTORY_CHATS_RESTORED")
end

function M:RestoreChat(chat)
	if chat.type == M.Enum.ChatType.BattleNet then
		local friend = Social:GetFriendByBattleTag(chat.battleTag)

		if friend then
			chat.accountID, chat.battleTag, chat.unit = friend.accountID, friend.battleTag, friend.unit
		else
			chat.accountID = nil
		end
	end

	chat.label = GetChatLabel(chat.battleTag, chat.unit)
end

function M:CopyChat(chat)
	local lines = {}

	for _, message in next, chat.messages do
		local copyColor = COLOR[chat.type][message.type]
		local copyPrefix = PREFIX[message.type]
		local copyText = COPY_FORMAT:format(copyPrefix, chat.label, message.text)
		local copyLine = copyColor:WrapTextInColorCode(copyText)

		table.insert(lines, copyLine)
	end

	self:SendMessage("SIXUI_COPY_LINES", lines)
end

function M:CreateChat(chatType, accountID, battleTag, unit)
	local chat = setmetatable({
		[M.Enum.Chat.AccountId] = accountID,
		[M.Enum.Chat.BattleTag] = battleTag,
		[M.Enum.Chat.Label] = GetChatLabel(battleTag, unit),
		[M.Enum.Chat.Messages] = {},
		[M.Enum.Chat.Type] = chatType,
		[M.Enum.Chat.Unit] = unit,
	}, CHAT_METATABLE)

	self:SendMessage("SIXUI_HISTORY_CHAT_CREATED", chat)

	return chat
end

function M:CreateMessage(chatType, accountID, battleTag, unit, messageType, text)
	local chat = self:FindOrCreateChat(chatType, accountID, battleTag, unit)

	local message = setmetatable({
		[M.Enum.Message.Read] = (messageType == M.Enum.MessageType.Sent) and GetServerTime(),
		[M.Enum.Message.Text] = text,
		[M.Enum.Message.Time] = GetServerTime(),
		[M.Enum.Message.Type] = messageType,
	}, MESSAGE_METATABLE)

	self:SendMessage("SIXUI_HISTORY_MESSAGE_CREATED", chat, message)

	return message
end

function M:FindChat(chatType, accountID, battleTag, unit)
	local isBattleNet = (chatType == M.Enum.ChatType.BattleNet)
	local isCharacter = (chatType == M.Enum.ChatType.Character)

	for chatIndex, chat in next, chats do
		if chat.type == chatType then
			if (isBattleNet and chat.battleTag == battleTag) or (isCharacter and chat.unit == unit) then
				return chat
			end
		end
	end
end

function M:FindOrCreateChat(...)
	return self:FindChat(...) or self:CreateChat(...)
end

function M:DeleteChat(chatIndex)
	table.remove(chats, chatIndex)

	self:SendMessage("SIXUI_HISTORY_CHAT_DELETED")
end

function M:MoveToBottom(chatIndex)
	if chats[chatIndex] then
		table.insert(chats, #chats, table.remove(chats, chatIndex))
	end
end

function M:MoveToTop(chatIndex)
	if chats[chatIndex] then
		table.insert(chats, 1, table.remove(chats, chatIndex))
	end
end

function M:SortChats()
	table.sort(chats)
end

function M:TruncateChats(limit)
	table.sort(chats)

	while #chats > limit do
		self:DeleteChat(#chats)
	end
end

function M:TruncateMessages(messages)
	while #messages > 25 do
		table.remove(messages, 1)
	end
end

-------------------------------------------------------------------------------
-- DataObject
-------------------------------------------------------------------------------

function O.OnEnter(frame)
	if frame then
		Tooltip:UpdateConfig({anchor = frame})
	end

	Tooltip:Show()
end

function O.OnClick(frame, button)
	if chats[1] then
		A:HandleClick({
			{mouseButton = C.Enum.MouseButton.Left,   modifierKey = C.Enum.ModifierKey.Any,   callback = function () Chat_OnClick(nil, 1, button) end},
			{mouseButton = C.Enum.MouseButton.Middle, modifierKey = C.Enum.ModifierKey.None,  callback = function () M:SortChats(); Tooltip:Refresh() end},
			{mouseButton = C.Enum.MouseButton.Right,  modifierKey = C.Enum.ModifierKey.None,  callback = function () showingList = not showingList; if showingList then table.sort(chats) end; Tooltip:Refresh() end},
			{mouseButton = C.Enum.MouseButton.Right,  modifierKey = C.Enum.ModifierKey.Shift, callback = function () M:CopyChat(chats[1]) end},
		}, button)
	end
end

function O.OnMouseWheel(frame, value)
	if not showingList then
		if value > 0 then
			M:MoveToTop(#chats)
		else
			M:MoveToBottom(1)
		end

		Tooltip:Refresh()
	end
end

function O:Update()
	local chat = chats[1]
	local chatType = chat and chat.type
	local message = chat and chat.lastMessage
	local messageType = message and message.type
	local color = chat and message and COLOR[chatType][messageType]

	if not chat or not message then
		-- Empty
		self.text = ""
		self.glyph = C.Enum.Glyph.CommentSlash
	elseif not message.read then
		-- Unread
		self.text = color:WrapTextInColorCode(chat.label)
		self.glyph = color:WrapTextInColorCode(C.Enum.Glyph.Comment)
	else
		-- Normal
		self.text = chat.label
		self.glyph = C.Enum.Glyph.Comment
	end
end

-------------------------------------------------------------------------------
-- Tooltip
-------------------------------------------------------------------------------

function Tooltip:OnShow()
	if not Tooltip:IsShown() then
		showingList = false
	end

	if not chats[1] then
		self:AddHeader(HISTORY)
		self:AddSeparator(C.TOOLTIP_SEPARATOR_HEIGHT)
		self:AddLine(ERR_NO_REPLY_TARGET)
	elseif showingList then
		self:AddChats()
		self:AddClearAll()
	else
		self:AddChat(chats[1])
	end
end

function Tooltip:AddChats()
	for chatIndex, chat in next, chats do
		local lineColor = COLOR[chat.type][M.Enum.MessageType.Received]
		local lineText = lineColor:WrapTextInColorCode(chat.label)
		local line = self:AddLine(lineText)

		self:SetLineScript(line, "OnEnter", Chat_OnEnter, chatIndex)
		self:SetLineScript(line, "OnLeave", Chat_OnLeave)
		self:SetLineScript(line, "OnMouseUp", Chat_OnClick, chatIndex)
	end
end

function Tooltip:AddClearAll()
	self:AddSeparator(C.TOOLTIP_SEPARATOR_HEIGHT)
	self:SetLineScript(self:AddLine("|cffffffff" .. CLEAR_ALL .. "|r"), "OnMouseUp", ClearAll_OnClick)
end

function Tooltip:AddChat(chat)
	self:AddHeader(chat.label)
	self:AddSeparator(C.TOOLTIP_SEPARATOR_HEIGHT)

	for _, message in next, chat.messages do
		self:AddMessage(chat, message)
	end
end

function Tooltip:AddMessage(chat, message)
	local lineColor = COLOR[chat.type][message.type]
	local lineText = lineColor:WrapTextInColorCode(message.text)

	self:SetCell(self:AddLine(), 1, lineText, nil, nil, nil, nil, nil, nil, 475, 150)

	if not message.read then
		M:SendMessage("SIXUI_HISTORY_MESSAGE_READ", chat, message)
	end
end

-------------------------------------------------------------------------------
-- Tooltip2
-------------------------------------------------------------------------------

function Tooltip2:OnShow(anchor, chat)
	self:SetAutoHideDelay(0)
	self:SmartAnchorTo(anchor, 10, 6)

	for _, message in next, chat.messages do
		Tooltip.AddMessage(self, chat, message)
	end
end

-------------------------------------------------------------------------------
-- Events
-------------------------------------------------------------------------------

function M:CHAT_MSG_BN_WHISPER(event, text, ...)
	if A.IsFiltered(event, text, ...) then return end
	local friend = Social:GetFriendByAccountID(select(12, ...))
	if (not friend) then return end
	self:CreateMessage(CHAT_TYPE[event], friend.accountID, friend.battleTag, friend.unit, MESSAGE_TYPE[event], text)
end

function M:CHAT_MSG_WHISPER(event, text, playerName, ...)
	if A.IsFiltered(event, text, playerName, ...) then return end
	self:CreateMessage(CHAT_TYPE[event], nil, nil, A.FormatUnit(playerName, nil, true), MESSAGE_TYPE[event], text)
end
